class PyLRCTool(Exception):
    """Exceptions used by pylrctool."""


class LRCParseError(PyLRCTool):
    """Error parsing LRC data."""


class LRCUnsupportedType(PyLRCTool):
    """Unsupported LRC event type."""


class LCRMissingTime(PyLRCTool):
    """LRC event missing time information."""



from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from zlib import crc32

import pendulum

from .exceptions import LCRMissingTime, LRCParseError, LRCUnsupportedType
from .regex import LRC_LINE, TIME_TAG


@dataclass(repr=False, eq=False, order=False)
class LRCEvent:
    class Type(str, Enum):
        Comment = "#"  # Comments
        Title = "ti"  # Title of the song
        Artist = "ar"  # Artist performing the song
        Album = "al"  # Album the song is from
        Author = "au"  # Author of the song
        Lyricist = "lr"  # Lyricist of the song
        Length = "length"  # Length of the song (mm:ss)
        FileAuthor = "by"  # Author of the LRC file (not the song)
        Offset = "offset"  # Specifies a global offset value for the lyric times, in milliseconds
        Creator = "re"  # The player or editor that created the LRC file
        Version = "ve"  # The version of the program
        TimedLyric = "lyric"  # [mm:ss.xx]text

    time: pendulum.Duration | None = None
    repeat_time: pendulum.Duration | None = None
    data: str = ""
    type: Type = Type.TimedLyric

    @cached_property
    def is_lyric(self) -> bool:
        return self.type is self.Type.TimedLyric

    @property
    def id(self) -> str:
        """Compute an ID from the LCR data."""
        checksum = crc32(self.compose.encode("utf8"))
        return hex(checksum)

    @property
    def time_code(self) -> str:
        """
        Convert time to LRC time code format [mm:ss.xx].
        :raises LCRMissingTime: If time is not set.
        """
        if self.is_lyric:
            if self.time is None:
                raise LCRMissingTime("Time must be set to generate time code.")
            return self._convert_time_to_code(self.time)
        return ""

    @property
    def repeat_time_code(self) -> str | None:
        """
        Convert repeat time to LRC time code format [mm:ss.xx].
        :return: Repeat time code string in the format [mm:ss.xx] or None if not set.
        """
        if self.is_lyric:
            if self.repeat_time is None:
                return None
            return self._convert_time_to_code(self.repeat_time)
        return ""

    @property
    def compose(self) -> str:
        """
        Compose the LRC event into a string format.
        Returns:
            str: The composed LRC event string.
        :raises LCRMissingTime: If time is required but not set for certain types.
        :raises LRCUnsupportedType: If the type is not recognized.
        """
        if isinstance(self.type, self.Type):
            if self.type == self.Type.TimedLyric:
                if self.time is None:
                    raise LCRMissingTime("Time must be set for timed lyric.")
                if self.repeat_time:
                    return f"[{self.time_code}][{self.repeat_time_code}]{self.data}"
                return f"[{self.time_code}]{self.data}"
            else:
                return f"[{self.type.value}:{self.data}]"

        raise LRCUnsupportedType(f"Unknown LRCEvent type: {self.type}")

    @classmethod
    def from_string(cls, lrc_string: str) -> LRCEvent | None:
        """
        Parse a string to create an LRCEvent.
        """
        if not (line_re := re.match(LRC_LINE, lrc_string)):
            raise LRCParseError(f"Invalid LRC event format: {lrc_string}")

        gd = line_re.groupdict()

        # Handle ID tags like [ti:Title]
        if tag := gd.get("tag"):
            return cls(time=None, data=f"{gd['value'].strip()}", type=cls.Type(tag))

        # Handle offset
        if gd.get("offset"):
            return cls(time=None, data=gd["offset"].strip(), type=cls.Type.Offset)

        # Handle comments
        if gd.get("comment"):
            return cls(time=None, data=gd["comment"].strip(), type=cls.Type.Comment)

        if gd.get("lyric") is not None:
            text = gd["lyric"].strip()
            timestamps = re.findall(TIME_TAG, lrc_string)

            time = None
            repeat_time = None

            time = pendulum.duration(
                minutes=int(timestamps[0][0]),
                seconds=int(timestamps[0][1]),
                milliseconds=int(timestamps[0][2] or 0) * 10,
            )

            if len(timestamps) > 1:
                # TODO: Handle repeat time if it can be multiple
                repeat_time = pendulum.duration(
                    minutes=int(timestamps[1][0]),
                    seconds=int(timestamps[1][1]),
                    milliseconds=int(timestamps[1][2] or 0) * 10,
                )

            return cls(time=time, repeat_time=repeat_time, data=text)

        raise LRCParseError(f"Unrecognized LRC event: {lrc_string}")

    def _convert_time_to_code(self, time: pendulum.Duration) -> str:
        """
        Convert a pendulum Duration to LRC time code format [mm:ss.xx].
        :param time: pendulum Duration object.
        :return: Time code string in the format [mm:ss.xx].
        """
        total_seconds = time.total_seconds()
        minutes, seconds = divmod(total_seconds, 60)
        return f"{int(minutes):02}:{seconds:05.2f}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LRCEvent):
            if self.time:
                if not isinstance(other.time, pendulum.Duration):
                    return False

                return (
                    self.time.total_seconds() == other.time.total_seconds()
                    and self.repeat_time == other.repeat_time
                    and self.data == other.data
                    and self.type == other.type
                )
            return self.data == other.data and self.type == other.type
        else:
            raise TypeError("Cannot compare to non-LRCEvent object")

    def __ne__(self, other: object) -> bool:
        eq_result = self.__eq__(other)
        if eq_result is NotImplemented:
            return NotImplemented
        return not eq_result

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, LRCEvent):
            return NotImplemented
        if self.time is None or other.time is None:
            return False
        return self.time < other.time

    def __le__(self, other: object) -> bool:
        if not isinstance(other, LRCEvent):
            return NotImplemented
        if self.time is None or other.time is None:
            return False
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, LRCEvent):
            return NotImplemented
        if self.time is None or other.time is None:
            return False
        return not self.__le__(other)

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, LRCEvent):
            return NotImplemented
        if self.time is None or other.time is None:
            return False
        return not self.__lt__(other)

    def __repr__(self) -> str:
        if self.is_lyric:
            return f"<LRCEvent time={self.time_code} repeat_time={self.repeat_time_code} text={self.data!r} type={self.type.name}>"
        else:
            return f"<LRCEvent text={self.data!r} type={self.type.name}>"



from __future__ import annotations

from collections.abc import Iterable
from datetime import timedelta
from pathlib import Path
from zlib import crc32

import pendulum
import srt
from sortedcontainers import SortedKeyList

from .lrcevent import LRCEvent


class LRCFile(SortedKeyList):
    def __init__(self, iterable: Iterable[LRCEvent] | None = None):
        super().__init__(key=lambda event: event.time or pendulum.duration())
        for event in iterable or []:
            self.add(event)

    @property
    def id(self) -> str:
        """Compute an ID from the LRC data."""
        checksum = crc32("\n".join([event.id for event in self]).encode("utf8"))

        return hex(checksum)

    def add(self, value: LRCEvent) -> None:
        """
        Add an LRCEvent to the file, ensuring no duplicates.
        :param value: LRCEvent to add.
        :raises TypeError: If the value is not an LRCEvent.
        :return: None
        """
        if not isinstance(value, LRCEvent):
            raise TypeError(f"Can only add LRCEvent objects, not {type(value)}")

        if any((ev == value) for ev in self):
            return

        super().add(value)

    def dump_to_srt(self, path: Path | str) -> int:
        """
        Export LRC Data to SRT file.
        :param path: Path to save the SRT file.
        :return: Number of characters written to the file.
        """
        if isinstance(path, str):
            path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        srt_data = []
        data_len = len(self)
        index = 1
        for event in self:
            if not event.is_lyric:
                index += 1
                continue

            start = timedelta(seconds=event.time.total_seconds())

            repeat_time_end = None
            if event.repeat_time:
                repeat_time_end = timedelta(seconds=event.repeat_time.total_seconds())

            if index < data_len:
                end = timedelta(seconds=self[index].time.total_seconds())
            else:
                end = start + timedelta(seconds=5)

            srt_line = srt.Subtitle(
                index=index + 1,
                content=event.data,
                start=start,
                end=repeat_time_end or end,
            )

            srt_data.append(srt_line)

            if repeat_time_end:
                srt_line.start = repeat_time_end
                srt_line.end = end
                srt_line.end = end
                index += 1
                srt_line.index = index
                srt_data.append(srt_line)

            index += 1

        return path.write_text(srt.compose(srt_data), encoding="utf8")

    def dumps(self, data_only: bool = False) -> str:
        """
        Serialize the LRC data to a string.
        :return: Serialized LRC data as a string.
        """

        if data_only:
            lines = [
                event.data
                for event in self
                if event.is_lyric
                for _ in (1, 2)
                if (_ == 1 or event.repeat_time) and event.data
            ]
            return "\n".join(lines)
        return "\n".join(event.compose for event in self)

    def dump(self, path: Path | str, data_only: bool = False) -> int:
        """
        Save the LRC data to a file.
        :param path: Path to save the LRC file.
        :return: Number of characters written to the file.
        """
        if isinstance(path, str):
            path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        lrc_text = self.dumps(data_only)

        return path.write_text(lrc_text, encoding="utf8")

    def from_string(self, data: str) -> bool:
        """
        Load LRC data from a file.
        :param data: LRC data.
        :return: True if anything was added
        """
        for line in data.splitlines():
            if line:
                if event := LRCEvent.from_string(line.strip()):
                    self.add(event)

        return bool(self)

    def from_file(self, path: Path | str) -> bool:
        """
        Load LRC data from lrc file.
        :param path: Path of the LRC file.
        :return: True if anything was added
        """
        if isinstance(path, str):
            path = Path(path)

        if not path.exists():
            raise FileNotFoundError(f"LRC file not found: {path}")

        with path.open("r", encoding="utf8") as file:
            for line in file:
                if event := LRCEvent.from_string(line.strip()):
                    self.add(event)

        return bool(self)

    def from_srt(self, path: Path | str) -> bool:
        """
        Load LRC data from srt file.
        :param path: Path of the SRT file.
        :return: True if anything was added
        """
        if isinstance(path, str):
            path = Path(path)

        if not path.exists():
            raise FileNotFoundError(f"SRT file not found: {path}")

        for line in srt.parse(path.read_text()):
            self.add(
                LRCEvent(
                    time=pendulum.duration(
                        seconds=line.start.total_seconds(),
                    ),
                    data=line.content,
                )
            )

        return bool(self)

    def merge_lines(self) -> bool:
        """
        Merge 2 repeating lines leveraging repeat_time.
        :return: True if there was any merge
        """
        merged = False
        i = 1
        for event in self:
            if event.is_lyric:
                if i < len(self) and event.data == self[i].data:
                    event.repeat_time = self[i].time
                    merged = True
                    del self[i]
                i += 1

        return merged

    def shift(self, minutes: int = 0, seconds: int = 0, milliseconds: int = 0) -> None:
        """Shift line by the set time
        :param minutes: Minutes to shift.
        :param seconds: Seconds to shift.
        :param milliseconds: Milliseconds to shift.
        :return: None
        """

        shift_time = pendulum.duration(
            minutes=minutes, seconds=seconds, milliseconds=milliseconds
        )

        for event in self:
            event.time += shift_time

    def __repr__(self) -> str:
        return f"LRCFile({list(self)})"



# --- ID TAGS ---
TAG_NAME = r"(?:ti|ar|al|au|lr|length|by|offset|re|tool|ve)"
TAG_VALUE = r"[^\]]*"
ID_TAG = rf"^\[(?P<tag>{TAG_NAME}):(?P<value>{TAG_VALUE})\]$"

# --- SPECIAL CASES ---
OFFSET_TAG = r"^\[offset:(?P<offset>[+-]?\d+)\]$"
LENGTH_TAG = r"^\[length:\s*(?P<minutes>\d{{1,2}}):(?P<seconds>\d{{2}})\]$"

# --- COMMENTS ---
COMMENT_TAG = r"^\#(?P<comment>.*)$"

# --- TIMESTAMPS ---
TIME_TAG = r"\[(?P<minutes>\d{2}):(?P<seconds>\d{2})(?:\.(?P<centiseconds>\d{1,2}))?\]"
TIME_TAG_MULTI = rf"(?:{TIME_TAG})+"

# --- TIMED LYRICS ---
TIMED_LINE = rf"^{TIME_TAG_MULTI}(?P<lyric>.*)$"

# --- COMBINED MATCH ---
LRC_LINE = rf"(?:{ID_TAG}|{COMMENT_TAG}|{TIMED_LINE})"
